#include "dzt.h"

enum {
	TARGET_STRING,
	TARGET_GNOME_ICON_LIST,
	TARGET_MOZ_URL,
	TARGET_URI_LIST,

	TARGET_OTHER
};

typedef gchar *(*DecodeFunc) (guchar *data, gint len, gchar *type);

static gchar *dnd_decode_string		(guchar *data, gint len);
static gchar *dnd_decode_gnome_icon_list(guchar *data, gint len);
static gchar *dnd_decode_moz_url	(guchar *data, gint len);
static gchar *dnd_decode_uri_list	(guchar *data, gint len);
static gchar *dnd_decode_other		(guchar *data, gint len, gchar *type);

static GtkTargetEntry dnd_targets[] = {
	{ "STRING",			0, TARGET_STRING },
	{ "x-special/gnome-icon-list",	0, TARGET_GNOME_ICON_LIST },
	{ "text/x-moz-url",		0, TARGET_MOZ_URL },
	{ "text/uri-list",		0, TARGET_URI_LIST },

	{ "",				0, TARGET_OTHER }
};

static gint dnd_ntargets = sizeof(dnd_targets)/sizeof(dnd_targets[0]);

static DecodeFunc dnd_decode_func[] = {
	(DecodeFunc)dnd_decode_string,
	(DecodeFunc)dnd_decode_gnome_icon_list,
	(DecodeFunc)dnd_decode_moz_url,
	(DecodeFunc)dnd_decode_uri_list,

	(DecodeFunc)dnd_decode_other
};

/* FIXME: Figure out *exactly* how targets work */

#ifdef TARGETS_WORK
GtkTargetEntry *
dnd_get_targets(gint *count)
{
	*count = dnd_ntargets;
	return dnd_targets;
}

gchar *
dnd_decode(GtkSelectionData *data, guint info)
{
	return dnd_decode_func[info] (data->data, data->length, NULL);
}

#else /* !TARGETS_WORK */

GtkTargetEntry *
dnd_get_targets(gint *count) 
{
	*count = 0;
	return NULL;
}

gchar *
dnd_decode(GtkSelectionData *data, guint info)
{
	gchar *type = gdk_atom_name(data->type);

	info = 0;
	while (info < dnd_ntargets - 1) {
		if (!strcmp(type, dnd_targets[info].target))
			break;
		info++;
	}

	info = dnd_targets[info].info;
	return dnd_decode_func[info] (data->data, data->length, type);
}
#endif

static gchar *
dnd_decode_string(guchar *data, gint len)
{
	return g_strdup(data);
}

static gchar *
dnd_decode_gnome_icon_list(guchar *data, gint len)
{
	return dnd_decode_uri_list(data, len);
}

static gchar *
dnd_decode_moz_url(guchar *data, gint len)
{
	gchar *ret = g_new0(gchar, len/2 + 1);
	gint i;

	for (i=0; i<len && data[i] != 0x0a ; i += 2)
		ret[i/2] = (gchar)data[i];

	return ret;
}

static gchar *
dnd_decode_uri_list(guchar *data, gint len)
{
	static gchar delimiters[] = { 0x0d, 0x00 };
	static gchar junk[]       = { 0x0a, 0x00 };

	GString *str	= g_string_sized_new(256);
	gchar   **elems = NULL;
	gchar	**elem	= NULL;
	gchar   *ret    = NULL;

	data  = g_strdelimit(data, junk, 0x0d);
	elems = g_strsplit(data, delimiters, 256);

	for (elem = elems; *elem; elem ++) {
		if (strncmp(*elem, "file://", 7))
			continue;

		g_string_sprintfa(str, "%s ", *elem + 7);
	}

	g_strfreev(elems);

	ret = str->str;
	g_string_free(str, FALSE);

	return ret;
}

static gchar *
dnd_decode_other(guchar *data, gint len, gchar *type)
{
	if (GET_BOOL(main.debug_dnd)) {
		GString *string = g_string_sized_new(256);
		gchar *ret;
		gint i;

		g_string_sprintfa(string, "\nUNKNOWN DATA TYPE: %s\n", type);
		for (i=0; i<len; i += 10) {
			gint j, l = i+10;
	
			for (j=i; j<l && j<len; j++)
				g_string_sprintfa(string, "0x%.2x ", data[j]);
	
			while (j++ < l)
				string = g_string_append(string, "     ");
		
			string = g_string_append(string, " \t ");
	
			for (j=i; j<l && j<len; j++) {
				guchar c = isgraph(data[j]) ? data[j] : '.';
				string = g_string_append_c(string, c);
			}
			string = g_string_append_c(string, '\n');
		}
	
		ret = string->str;
		g_string_free(string, FALSE);

		gdk_beep();

		return ret;
	}
	return NULL;
}
