/*
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
/*
 * Copyright (c)2001,2002 Adrian Drzewiecki
 */
#include "dzt.h"
#include "prefs-gui.h"

gchar dzt_program [] = PACKAGE;
gchar dzt_version [] = VERSION;

static void show_error(gchar *error);
static gboolean init_config(void);
static gboolean check_config_changed(void);

int main(int argc, char ** argv)
{
	gint noremote = FALSE;
	gchar *error = NULL;

	struct poptOption options [] = {
		{
			"disable-remote",
			'r',
			POPT_ARG_NONE,
			&noremote,
			0,
			N_("Disable remote terminal capabilities"),
			NULL
		},
				{
			"error",
			'e',
			POPT_ARG_STRING,
			&error,
			0,
			N_("Show an error dialog box."),
			N_("ERROR TEXT")
		},
		{
			NULL,
			'\0',
			0,
			NULL,
			0,
			NULL,
			NULL
		}
	};

	poptContext pctx;

	gnome_init_with_popt_table(dzt_program, dzt_version,
					argc, argv, options, 0, &pctx);

	if (error)
		show_error(error);

	if (!noremote && remote_init() && remote_start(""))
		gtk_exit(0);

	if (init_config())
		config_sync();

	window_create(AUTOSTART);

	gtk_main();
	return 0;
}

void
config_sync()
{
	time_t now = time(NULL);
	struct utimbuf buf = { now, now };
	gnome_config_sync();
	utime(gnome_config_get_real_path(""), &buf);
}

gboolean
config_changed()
{
	static struct stat config_stat;
        static struct stat old_config_stat;
        static gchar *filename = NULL;

        if(filename == NULL) {
                filename = gnome_config_get_real_path("");
                stat(filename, &old_config_stat);
        } else {
                stat(filename, &config_stat);
                if((config_stat.st_mtime == old_config_stat.st_mtime) &&
                   (config_stat.st_ctime == old_config_stat.st_ctime))
                        return FALSE;
                memcpy(&old_config_stat, &config_stat, sizeof(struct stat));
        }

	return TRUE;
}

static void
show_error(gchar *error)
{
	GtkWidget *mb = gnome_message_box_new(error,
						GNOME_MESSAGE_BOX_ERROR,
						GNOME_STOCK_BUTTON_OK,
						NULL);

	gnome_dialog_button_connect(GNOME_DIALOG(mb), 0,
			GTK_SIGNAL_FUNC(gtk_exit), 0);

	gtk_window_set_position(GTK_WINDOW(mb), GTK_WIN_POS_MOUSE);

	gtk_widget_show(mb);

	gtk_main();
}

static gboolean
init_config()
{
	gboolean initialized;
	gnome_config_push_prefix(DZT_CONFIG_PREFIX "prefs/");
	initialized = push_profile_id(DEFAULT_PROFILE_ID);
	if (!initialized) {
		SET_BOOL(main.show_shortcuts, TRUE);
		SET_BOOL(main.no_right_click_on_terminal, FALSE);
		SET_INT(main.toolbar_style, GTK_TOOLBAR_BOTH);
		SET_BOOL(main.confirm_close_all, TRUE);

		SET_BOOL(notebook.expand_tabs, FALSE);
		SET_INT(notebook.tab_position, GTK_POS_TOP);
		SET_BOOL(notebook.single_terminal_has_tab, TRUE);
		SET_INT(notebook.max_title_length, 64);
		SET_INT(notebook.title_padding, 2);
		SET_BOOL(notebook.tabs_have_close_button, TRUE);

		SET_STR(shortcuts.menu, "");
		SET_STR(shortcuts.detach_tab, "");
		SET_STR(shortcuts.rename_tab, "");
		SET_STR(shortcuts.close_tab, "");
		SET_STR(shortcuts.prev_tab, "");
		SET_STR(shortcuts.next_tab, "");
		SET_STR(shortcuts.close_win, "");
		SET_STR(shortcuts.quit, "");
		SET_STR(shortcuts.prefs, "");

		push_new_profile();
		SET_STR(program.name, "Default");
		SET_STR(program.command, "$SHELL");
		SET_INT(program.kill_signal, 0);
		SET_BOOL(program.restart, FALSE);
		SET_BOOL(program.change_notify, TRUE);
		SET_STR(program.new_tab_key, "");
		SET_STR(program.new_win_key, "");
		SET_BOOL(program.autostart, FALSE);

		SET_INT(terminal.background_type, COLOR);
		SET_STR(terminal.background_image, "");
		SET_STR(terminal.type, "xterm");
		SET_STR(terminal.word_chars, "-A-Za-z0-9/_:.,?+%=");
		SET_STR(terminal.font, "-misc-fixed-*-*-*-*-*-*-*-*-*-*-*-*");
		SET_BOOL(termianl.enable_bell, TRUE);
		SET_BOOL(terminal.del_is_del, TRUE);
		SET_BOOL(terminal.blinking_cursor, TRUE);
		SET_BOOL(terminal.update_title, TRUE);
		SET_BOOL(termianl.swap_del, FALSE);

		SET_INT(scrolling.scrollbar_location, BOTH);
		SET_INT(scrolling.scroll_buffer, 1000);
		SET_BOOL(scrolling.scroll_on_keystroke, TRUE);
		SET_BOOL(scrolling.scroll_on_output, FALSE);

		SET_STR(colors.foreground, "0x0000 0x0000 0x0000");
		SET_STR(colors.background, "0xeeee 0xeeee 0xeeee");
		SET_STR(colors.pal0, "0x5555 0x5555 0x5555");
		SET_STR(colors.pal1, "0xffff 0x5555 0x5555");
		SET_STR(colors.pal2, "0x5555 0xffff 0x5555");
		SET_STR(colors.pal3, "0xffff 0xffff 0x5555");
		SET_STR(colors.pal4, "0x5555 0x5555 0xffff");
		SET_STR(colors.pal5, "0xffff 0x5555 0xffff");
		SET_STR(colors.pal6, "0x5555 0xffff 0xffff");
		SET_STR(colors.pal7, "0xffff 0xffff 0xffff");
		SET_STR(colors.pal8, "0x0000 0x0000 0x0000");
		SET_STR(colors.pal9, "0xaaaa 0x0000 0x0000");
		SET_STR(colors.pal10, "0x0000 0xaaaa 0x0000");
		SET_STR(colors.pal11, "0xaaaa 0x5555 0x0000");
		SET_STR(colors.pal12, "0x0000 0x0000 0xaaaa");
		SET_STR(colors.pal13, "0xaaaa 0x0000 0xaaaa");
		SET_STR(colors.pal14, "0x0000 0xaaaa 0xaaaa");
		SET_STR(colors.pal15, "0xaaaa 0xaaaa 0xaaaa");
		SET_STR(colors.change_notify, "0xffff 0x0000 0x0000");
		SET_BOOL(colors.invert, FALSE);
	}
	pop_profile();

	gtk_timeout_add(300, (GtkFunction)check_config_changed, NULL);

	return !initialized;
}

static gboolean
check_config_changed()
{
	if (config_changed()) {
		menu_configure();
		window_configure_all();
		update_preferences_dialog();
	}

	return TRUE;
}
