/*
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
/*
 * Copyright (c)2001,2002 Adrian Drzewiecki
 */
#include"../config.h"

#include<gtk/gtk.h>
#include<string.h>
#include<errno.h>
#include<stdio.h>
#include<signal.h>
#include<pwd.h>
#include<unistd.h>
#include<sys/file.h>
#include<sys/types.h>
#include<sys/stat.h>
#include<gdk/gdkkeysyms.h>
#include<gnome.h>
#include<zvt/zvtterm.h>
#include<ctype.h>
#include<fcntl.h>
#include<utime.h>

/* terminal background types */
#define COLOR			0
#define IMAGE			1
#define SCROLLING_IMAGE		2
#define TRANSPARENT		3
#define TRANSPARENT_SHADED	4

/* scrollbar positions */
#define LEFT	0
#define RIGHT	1
#define BOTH	2
#define NONE	3

/* dzt misc methods */
gboolean	config_changed			(void);
void		config_sync			(void);
void		update_preferences_dialog	(void);
void		edit_prefs			(void);

/* page methods */
GtkWidget *	page_create		(GtkWidget *notebook, gint id);
void		page_begin_rename	(GtkWidget *page, gboolean steal_focus);
void		page_end_rename		(GtkWidget *page, gint ok);
void		page_rename		(GtkWidget *page, gchar *name);
void		page_update_title	(GtkWidget *page);
void		page_grab_focus		(GtkWidget *page);
void		page_configure		(GtkWidget *page);
void		page_move_left		(GtkWidget *page);
void		page_move_right		(GtkWidget *page);
void		page_moveto		(GtkWidget *page, GtkWidget *notebook);
gint		page_is_lonely		(GtkWidget *page);
void		page_remove		(GtkWidget *page);

/* notebook methods */
GtkWidget *	notebook_create		(void);
GtkWidget *	notebook_current_page	(GtkWidget *notebook);
void		notebook_update_title	(GtkWidget *notebook, gchar *title);
void		notebook_configure	(GtkWidget *notebook);
void		notebook_remove_page	(GtkWidget *notebook, GtkWidget *page);
void		notebook_set_ignore_click (GtkWidget *notebook);
void		notebook_grab_focus	(GtkWidget *notebook);
void		notebook_go_prev_page	(GtkWidget *notebook);
void		notebook_go_next_page	(GtkWidget *notebook);
void		notebook_close		(GtkWidget *notebook);


/* menu methods */
void		menu_configure		(void);
void		menu_popup		(GtkWidget *notebook, guint32 when);
gint		menu_do_accel		(GtkWidget *notebook, guint stete, guint keyval);

/* window methods */
void		window_create		(gint id);
void		window_create_with_page	(GtkWidget *page);
void		window_update_title	(GtkWidget *window, gchar *title);
void		window_configure	(GtkWidget *window);
void		window_configure_all	(void);
void		window_close		(GtkWidget *window);
void		window_close_all	(void);

/* remote control methods */
gint		remote_init		(void);
gint		remote_start		(gchar *prog);

/* command-line parsing methods */
gchar **	program_parse_command	(gchar *command);

/* dnd methods */
GtkTargetEntry *dnd_get_targets		(gint *count);
gchar *		dnd_decode		(GtkSelectionData *data, guint info);

/* object data access */
#define GET_OBJECT_INFO(obj, InfoType) \
	((InfoType*)get_object_info(GTK_OBJECT(obj), "$" #InfoType, sizeof(InfoType)))

static inline gpointer
get_object_info(GtkObject *__obj, gchar *__name, gint __size)
{
	gpointer *__info = gtk_object_get_data(__obj, __name);
	if (!__info) {
		__info = g_malloc0(__size);
		g_assert(__info != NULL);
		gtk_object_set_data_full(__obj, __name, __info, (GtkDestroyNotify)g_free);
	}
	return __info;
}

#define MOVE_OBJECT_INFO(from, to, InfoType) \
	move_object_info(GTK_OBJECT(from), GTK_OBJECT(to), "$" #InfoType, sizeof(InfoType))

static inline void
move_object_info(GtkObject *__from, GtkObject *__to, gchar *__name, gint __size)
{
	gpointer *__from_info = get_object_info(__from, __name, __size);
	gpointer *__to_info   = get_object_info(__to, __name, __size);

	memcpy(__to_info, __from_info, __size);

	gtk_object_remove_data(__from, __name);
}

/* config helpers */
#define DZT_CONFIG_PREFIX		"/dzt2/"

#define GET_STR(__path)			gnome_config_get_string(#__path)
#define GET_INT(__path)			gnome_config_get_int(#__path)
#define GET_BOOL(__path)		gnome_config_get_bool(#__path)
#define GET_COLOR(__path)		get_color_real(#__path)
#define GET_NTH_COLOR(__path, __n)	get_nth_color_real(#__path, __n)

static inline gushort *
get_color_real(gchar *path)
{
	gchar *value = gnome_config_get_string(path);
	gushort *color = NULL;
	if (value) {
		gchar *ptr = value;
		color = g_new0(gushort, 3);
		color[0] = strtol(ptr, &ptr, 0);
		color[1] = strtol(ptr, &ptr, 0);
		color[2] = strtol(ptr, &ptr, 0);
		g_free(value);
	}
	return color;
}

static inline gushort *
get_nth_color_real(gchar *path, gint n)
{
	gchar *full_path = g_strdup_printf("%s%d", path, n);
	gushort *color = get_color_real(full_path);
	g_free(full_path);
	return color;
}

#define SET_STR(__path, __value)		gnome_config_set_string(#__path, __value)
#define SET_INT(__path, __value)		gnome_config_set_int(#__path, __value)
#define SET_BOOL(__path, __value)		gnome_config_set_bool(#__path, __value)
#define SET_COLOR(__path, __value)		set_color_real(#__path, __value)
#define SET_NTH_COLOR(__path, __n, __value)	set_nth_color_real(#__path, __n, __value)

static inline void
set_color_real(gchar *path, gushort *value)
{
	gchar *real_value = g_strdup_printf("0x%x 0x%x 0x%x",
					value[0], value[1], value[2]);
	gnome_config_set_string(path, real_value);
	g_free(real_value);
}

static inline void
set_nth_color_real(gchar *path, gint n, gushort *value)
{
	gchar *full_path = g_strdup_printf("%s%d", path, n);
	set_color_real(full_path, value);
	g_free(full_path);
}

/* profiles */
#define AUTOSTART		0
#define DEFAULT_PROFILE_ID	1

static inline const gchar *
profile_path(gint nr)
{
	static gchar path[PATH_MAX];

	g_snprintf(path, sizeof(path) - 1, DZT_CONFIG_PREFIX "profile%d/", nr);
	if (!gnome_config_has_section(path))
		return NULL;

	return path;
}

static inline void
pop_profile()
{
	gnome_config_pop_prefix();
}

static inline gboolean
push_profile_nr(gint nr)
{
	const gchar *path = profile_path(nr);
	if (!path)
		return FALSE;
	gnome_config_push_prefix(path);
	return TRUE;
}

static inline gint
find_profile_nr(gint id)
{
	gint nr;

	for(nr = 0; push_profile_nr(nr); nr ++) {
		if (id == GET_INT(id)) {
			pop_profile();
			return nr;
		}
		pop_profile();
	}

	return 0;
}

static inline gint
get_profile_id(gint nr)
{
	gint id = DEFAULT_PROFILE_ID;

	if (push_profile_nr(nr)) {
		id = GET_INT(id);
		pop_profile();
	}

	return id;
}

static inline gboolean
push_profile_id(gint id)
{
	return push_profile_nr(find_profile_nr(id));
}

static inline int
push_new_profile()
{
	gchar path[PATH_MAX];
	gint id = DEFAULT_PROFILE_ID;
	gint nr = 0;

	while(profile_path(nr) != NULL)
		nr ++;

	if (nr)
		id = (gint)time(NULL);

	g_snprintf(path, sizeof(path) - 1, DZT_CONFIG_PREFIX "profile%d/", nr);
	gnome_config_push_prefix(path);
	SET_INT(id, id);

	return nr;
}

static inline void
remove_profile_nr(gint nr)
{
	if (nr) {
		const gchar *path;
		while ((path = profile_path(nr + 1))) {
			gpointer iter;
			gchar *key;
			gchar *val;

			iter = gnome_config_init_iterator(path);
			push_profile_nr(nr);
			while((iter = gnome_config_iterator_next(iter, &key, &val))) {
				gnome_config_set_string(key, val);
				g_free(key);
				g_free(val);
			}
			pop_profile();

			nr ++;
		}

		gnome_config_clean_section(profile_path(nr));
	}
}

static inline void
swap_profile_nrs(gint nr1, gint nr2)
{
	if (nr1 && nr2 && profile_path(nr1) && profile_path(nr2)) {
		gpointer iter = gnome_config_init_iterator(profile_path(nr1));
		gchar *key;
		gchar *val;
		gchar *oldval;

		while((iter = gnome_config_iterator_next(iter, &key, &val))) {
			push_profile_nr(nr2);
			oldval = gnome_config_get_string(key);
			gnome_config_set_string(key, val);
			pop_profile();

			push_profile_nr(nr1);
			gnome_config_set_string(key, oldval);
			pop_profile();

			g_free(key);
			g_free(val);
			g_free(oldval);
		}
	}
}
