/*
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
/*
 * Copyright (c)2001,2002 Adrian Drzewiecki
 */
#include "dzt.h"

static GtkMenu *menu    = NULL;

/*
 * These two are kinda nasty hack
 * to get menus to display correctly
 * when using keyboard accelerators
 */
static GtkWidget *popup	  = NULL;
static gint      in_accel = FALSE;

static GtkWidget *current_notebook = NULL;

static GtkMenu *menu_configure_profiles_submenu (gchar *label,
						 gchar *accel,
					         gchar *icon,
						 GtkSignalFunc action);

static void menu_configure_item (gchar *label,
				 gchar *accel,
		                 gchar *icon,
			         GtkSignalFunc action);

static void menu_item_set_accel (GtkWidget *item, gchar *accel);

static void on_menu_new_tab	(GtkWidget *item, gint profileid);
static void on_menu_new_window	(GtkWidget *item, gint profileid);

static void on_menu_moveleft_tab	(void);
static void on_menu_moveright_tab	(void);

static void on_menu_detach_tab	(void);
static void on_menu_rename_tab	(void);
static void on_menu_close_tab	(void);
static void on_menu_prev_tab	(void);
static void on_menu_next_tab	(void);

static void on_menu_close_window	(void);
static void on_menu_close_all_windows	(void);

static void on_menu_prefs (void);

static void on_menu_popup(GtkWidget *item, GtkWidget *menu);

void
menu_configure()
{
	GtkWidget *item;
	gchar dzt [256];

	extern gchar dzt_program [];
	extern gchar dzt_version [];

	if (menu)
		gtk_widget_destroy(GTK_WIDGET(menu));

	menu = GTK_MENU(gtk_menu_new());

	snprintf(dzt, sizeof(dzt) - 1, "%s %s", dzt_program, dzt_version);

	item = gtk_menu_item_new_with_label(dzt);
	gtk_widget_set_sensitive(item, FALSE);
	gtk_menu_append(menu, item);

	if (popup)
		gtk_widget_destroy(popup);

	popup = gtk_menu_item_new();
	gtk_signal_connect(GTK_OBJECT(popup), "activate",
			GTK_SIGNAL_FUNC(on_menu_popup), (gpointer)menu);
	menu_item_set_accel(popup,
			GET_STR(shortcuts.menu));

	gtk_menu_append(menu, gtk_menu_item_new());

	menu_configure_profiles_submenu(
			"new tab",
			NULL,
			GNOME_STOCK_MENU_EXEC,
			on_menu_new_tab);

	menu_configure_profiles_submenu(
			"new window",
			NULL,
			GNOME_STOCK_MENU_NEW,
			on_menu_new_window);

	menu_configure_item(
			"detach tab",
			GET_STR(shortcuts.detach_tab),
			GNOME_STOCK_MENU_UNDO,
			GTK_SIGNAL_FUNC(on_menu_detach_tab));

	menu_configure_item(
			"rename tab",
			GET_STR(shortcuts.rename_tab),
			GNOME_STOCK_MENU_PREF,   
			GTK_SIGNAL_FUNC(on_menu_rename_tab));

	menu_configure_item(
			"close tab", 
			GET_STR(shortcuts.close_tab),
			GNOME_STOCK_MENU_CLOSE,  
			GTK_SIGNAL_FUNC(on_menu_close_tab));
	menu_configure_item(
			"move tab left",
			GET_STR(shortcuts.moveleft_tab),
			GNOME_STOCK_MENU_FIRST,
			GTK_SIGNAL_FUNC(on_menu_moveleft_tab));
	menu_configure_item(
			"move tab right",
			GET_STR(shortcuts.moveleft_tab),
			GNOME_STOCK_MENU_LAST,
			GTK_SIGNAL_FUNC(on_menu_moveright_tab));
	
	menu_configure_item(
			"prev tab",  
			GET_STR(shortcuts.prev_tab),
			GNOME_STOCK_MENU_BACK,   
			GTK_SIGNAL_FUNC(on_menu_prev_tab));

	menu_configure_item(
			"next tab",  
			GET_STR(shortcuts.next_tab),
			GNOME_STOCK_MENU_FORWARD,
			GTK_SIGNAL_FUNC(on_menu_next_tab));

	menu_configure_item(
			"close window",     
			GET_STR(shortcuts.close_win),
			GNOME_STOCK_MENU_STOP,
			GTK_SIGNAL_FUNC(on_menu_close_window));

	menu_configure_item(
			"close all windows",
			GET_STR(shortcuts.quit),
			GNOME_STOCK_MENU_QUIT,
			GTK_SIGNAL_FUNC(on_menu_close_all_windows));

	menu_configure_item(
			"preferences",
			GET_STR(shortcuts.prefs),
			GNOME_STOCK_MENU_PREF,
			GTK_SIGNAL_FUNC(on_menu_prefs));

	gtk_widget_show_all(GTK_WIDGET(menu));
	gtk_widget_hide(GTK_WIDGET(menu));
}

void
menu_popup(GtkWidget *notebook, guint32 when)
{
	current_notebook = notebook;
	gtk_menu_popup(menu, NULL, NULL, NULL, NULL, when, 0);
}

gint
menu_do_accel(GtkWidget *notebook, guint state, guint keyval)
{
	gint ok;
	current_notebook = notebook;
	in_accel = TRUE;
	ok = gtk_accel_groups_activate(GTK_OBJECT(notebook), keyval, state);
	in_accel = FALSE;
	return ok;
}

static GtkMenu *
menu_configure_profiles_submenu(gchar *label,
				gchar *accel,
		                gchar *icon,
				GtkSignalFunc action)
{
	GtkWidget *item = gnome_stock_menu_item(icon, _(label));
	GtkWidget *sub  = gtk_menu_new();
	gint profilenr = 0;

	gtk_object_set_data(GTK_OBJECT(sub), "action", (gpointer)action);
	
	gtk_menu_item_set_submenu(GTK_MENU_ITEM(item), sub);
	gtk_menu_append(menu, item);
	gtk_signal_connect(GTK_OBJECT(item), "activate",
			GTK_SIGNAL_FUNC(on_menu_popup), (gpointer)sub);
	menu_item_set_accel(item, accel);

	item = gtk_menu_item_new_with_label(label);
	gtk_widget_set_sensitive(item, FALSE);
	gtk_menu_append(GTK_MENU(sub), item);

	gtk_menu_append(GTK_MENU(sub), gtk_menu_item_new());

	while(push_profile_nr(profilenr++)) {
		label = GET_STR(program.name);
		item = gtk_menu_item_new_with_label(label);
		g_free(label);
		gtk_signal_connect(GTK_OBJECT(item),
				"activate",
				action,
				(gpointer)GET_INT(id));
		gtk_menu_append(GTK_MENU(sub), item);
		if (action == on_menu_new_window)
			menu_item_set_accel(item,
				GET_STR(program.new_win_key));
		else
			menu_item_set_accel(item,
				GET_STR(program.new_tab_key));

		pop_profile();
	}

	return GTK_MENU(sub);
}

static void
menu_configure_item(gchar *label,
		    gchar *accel, 
		    gchar *type,
		    GtkSignalFunc action)
{
	GtkWidget *item = gnome_stock_menu_item(type, _(label));
	gtk_signal_connect(GTK_OBJECT(item), "activate", action, NULL);
	gtk_menu_append(menu, item);
	menu_item_set_accel(item, accel);
}

static void
menu_item_set_accel(GtkWidget *item, gchar *accel)
{
	GtkAccelGroup   *group = gtk_accel_group_get_default();
	GdkModifierType mod;
	guint           key;
	gint            flags = 0;

	if (GET_BOOL(main.show_shortcuts))
		flags = GTK_ACCEL_VISIBLE;

	if (accel && strlen(accel)) {
		gtk_accelerator_parse(accel, &key, &mod);
		if (!gtk_accelerator_valid(key, mod))
			g_warning("Invalid accelerator '%s'.\n", accel);
		else
			gtk_accel_group_add(group, key, mod, flags,
					GTK_OBJECT(item), "activate");
	}
	g_free(accel);
}

static void
on_menu_new_tab(GtkWidget *item, gint profileid)
{
	page_create(current_notebook, profileid);
}

static void
on_menu_new_window(GtkWidget *item, gint profileid)
{
	window_create(profileid);
}

static void
on_menu_detach_tab()
{
	window_create_with_page(notebook_current_page(current_notebook));
}

static void
on_menu_rename_tab()
{
	page_begin_rename(notebook_current_page(current_notebook), FALSE);
}

static void
on_menu_close_tab()
{
	page_remove(notebook_current_page(current_notebook));
}

static void
on_menu_moveleft_tab()
{
	page_move_left(notebook_current_page(current_notebook));
}

static void
on_menu_moveright_tab()
{
	page_move_right(notebook_current_page(current_notebook));
}

static void
on_menu_prev_tab()
{
	notebook_go_prev_page(current_notebook);
}

static void
on_menu_next_tab()
{
	notebook_go_next_page(current_notebook);
}

static void
on_menu_close_window()
{
	notebook_close(current_notebook);
}

static void
on_menu_close_all_windows()
{
	window_close_all();
}

static void
on_menu_prefs()
{
	edit_prefs();
}

static void
on_menu_popup(GtkWidget *item, GtkWidget *menu)
{
	if (!menu->parent || in_accel)
		gtk_menu_popup(GTK_MENU(menu), NULL, NULL, NULL, NULL, 0, 0);
}
