/*
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
/*
 * Copyright (c)2001,2002 Adrian Drzewiecki
 */
#include "dzt.h"

typedef struct _PageInfo PageInfo;

struct _PageInfo
{
	GtkWidget *label;
	GtkWidget *entry;
	GtkWidget *zvt;
	GtkWidget *close_button;
	GtkWidget *lf_sbar;
	GtkWidget *rt_sbar;

	gint input_tag;

	gchar title[512];
	gchar command[512];

	gint profileid;

	gboolean restarting;
	gboolean entry_steal_focus; // hack!!
	gboolean user_title;
	gboolean changed;
};

#define page_info(page)			GET_OBJECT_INFO(page, PageInfo)
#define page_label(page)		(page_info(page)->label)
#define page_entry(page)		(page_info(page)->entry)
#define page_zvt(page)			(page_info(page)->zvt)
#define page_user_title(page)		(page_info(page)->user_title)
#define page_entry_steal_focus(page)	(page_info(page)->entry_steal_focus)
#define page_close_button(page)		(page_info(page)->close_button)
#define page_lf_sbar(page)		(page_info(page)->lf_sbar)
#define page_rt_sbar(page)		(page_info(page)->rt_sbar)
#define page_title(page)		(page_info(page)->title)
#define page_restarting(page)		(page_info(page)->restarting)
#define page_profileid(page)		(page_info(page)->profileid)
#define page_input_tag(page)		(page_info(page)->input_tag)
#define page_changed(page)		(page_info(page)->changed)

static inline GtkWidget *
page_notebook(GtkWidget *page)
{
	return gtk_widget_get_ancestor(page, GTK_TYPE_NOTEBOOK);
}

static inline GtkWidget *
page_from_widget(GtkWidget *w)
{
	while(w && !GTK_IS_NOTEBOOK(w->parent))
		w = w->parent;
	return w;
}

static inline GtkWidget *
page_from_label_component(GtkWidget *w)
{
	GtkWidget *notebook = gtk_widget_get_ancestor(w, GTK_TYPE_NOTEBOOK);
	GtkWidget *page;
	GtkWidget *cont;
	int pagenum = 0;

	w = gtk_widget_get_ancestor(w, GTK_TYPE_HBOX);
	for (;;) {
		page = gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook), pagenum ++);
		if (!page)
			return NULL;
		cont = gtk_notebook_get_tab_label(GTK_NOTEBOOK(notebook), page);
		if (cont == w)
			return page;
	}
}

static inline void
push_page_profile(GtkWidget *page)
{
	push_profile_id(page_profileid(page));
}

static GtkWidget * sbar_cont_new	(void);
static GtkWidget * page_create_helper	(GtkWidget *notebook, gint profileid);
static gboolean    page_start_command	(GtkWidget *page);

static void page_configure_terminal	(GtkWidget *page);
static void page_configure_scrolling	(GtkWidget *page);
static void page_configure_colors	(GtkWidget *page);
static void page_configure_notebook	(GtkWidget *page);

static void page_set_changed		(GtkWidget *page, gboolean value);

static gint on_page_destroy		(GtkWidget *page);
static void on_page_close_button_click	(GtkWidget *page);

static gint on_page_zvt_button_press	(GtkWidget *w, GdkEventButton *ev);
static gint on_page_sbar_button_press	(GtkWidget *w, GdkEventButton *ev);
static void on_page_zvt_child_died	(GtkWidget *w);

static void on_page_zvt_title_changed (GtkWidget *w,
		                       VTTITLE_TYPE type,
				       gchar *newtitle);

//static void on_page_zvt_realize	(GtkWidget *w);

static gboolean on_page_zvt_drag_motion (GtkWidget *zvt,
		     			 GdkDragContext *ctx,
					 gint x,
					 gint y,
					 guint time);

static gboolean on_page_zvt_drag_drop (GtkWidget *zvt,
				       GdkDragContext *ctx,
				       gint x,
				       gint y,
				       guint time);

static gboolean on_page_zvt_drag_data_received (GtkWidget *zvt,
				       		GdkDragContext *ctx,
				       		gint x,
				       		gint y,
						GtkSelectionData *data,
						guint info,
						guint time);

static void on_page_entry_keypress (GtkWidget *entry,
		                    GdkEventKey *event);

static void on_page_entry_focusout (GtkWidget *entry,
				    GdkEventFocus *event);

static gboolean on_page_restart_timeout (gpointer data);

static void on_page_zvt_output(gpointer data,
							   gint source,
							   GdkInputFunction condition);


GtkWidget *
page_create(GtkWidget *notebook, gint profileid)
{
	GtkWidget *page = page_create_helper(notebook, profileid);

	if (!page_start_command(page)) {
		page_remove(page);
		return NULL;
	} 

	return page;
}



void
page_begin_rename(GtkWidget *page, gboolean steal_focus)
{
	GtkWidget *label = page_label(page);
	GtkWidget *entry = page_entry(page);
	gchar *text;

	gtk_label_get(GTK_LABEL(label), &text);
	gtk_entry_set_text(GTK_ENTRY(entry), text);
	gtk_entry_set_editable(GTK_ENTRY(entry), TRUE);
	gtk_editable_select_region(GTK_EDITABLE(entry), 0, -1);

	gtk_widget_hide(label);
	gtk_widget_show(entry);

	page_entry_steal_focus(page) = steal_focus;
	gtk_widget_grab_focus(entry);
}

void
page_end_rename(GtkWidget *page, gint ok)
{
	GtkWidget *label = page_label(page);
	GtkWidget *entry = page_entry(page);
	gchar *name;

	name = gtk_entry_get_text(GTK_ENTRY(entry));
	gtk_entry_set_editable(GTK_ENTRY(entry), FALSE);
	gtk_widget_show(label);
	gtk_widget_hide(entry);

	if (strlen(name) && ok) {
		page_user_title(page) = FALSE;
		page_rename(page, name);
		page_user_title(page) = TRUE;
	} else
		gtk_widget_queue_resize(label->parent);
}

void
page_rename(GtkWidget *page, gchar *title)
{
	GtkWidget *label = page_label(page);
	GtkWidget *entry = page_entry(page);

	gchar *older;
	gchar *newer;
	gint len;
	gint max;
	gint end;

	if (!title || !strlen(title))
		return;

	gtk_label_get(GTK_LABEL(label), &older);
	if (older && strlen(older)) {
		push_page_profile(page);
		if (!GET_BOOL(terminal.update_title) ||
			(page_user_title(page) && GET_BOOL(terminal.user_title))) {
			pop_profile();
			return;
		}

		pop_profile();
	}

	if (title != page_title(page))
		g_snprintf(page_title(page), sizeof(page_title(page)) - 1, "%s", title);

	max = GET_INT(notebook.max_title_length);
	max = max < 10 ? 10 : max > 128 ? 128 : max;

	len = strlen(title);
	if (len < max) {
		newer = g_strdup(title);
	} else {
		end = 2 + len - max/2;
		len = max/2 - 4;
		newer = g_strdup_printf("%.*s...%s", len, title, (gchar*)(title + end));
	}

	if (!older || strcmp(older, newer)) {
		gtk_label_set_text(GTK_LABEL(label), newer);
		gtk_entry_set_text(GTK_ENTRY(entry), newer);
	}
	
	g_free(newer);

	page_update_title(page);

	gtk_widget_queue_resize(label->parent);
}

void
page_update_title(GtkWidget *page)
{
	GtkWidget *notebook = page_notebook(page);
	if (page == notebook_current_page(notebook))
		notebook_update_title(notebook, page_title(page));
}

void
page_grab_focus(GtkWidget *page)
{
	if (page) {
		if (page_entry_steal_focus(page)) {
			gtk_widget_grab_focus(page_entry(page));
			page_entry_steal_focus(page) = FALSE;
		
		} else if (page_zvt(page)) {
			page_update_title(page);
			gtk_widget_grab_focus(page_zvt(page));
			page_set_changed(page, FALSE);
		}
	}
}

void
page_configure(GtkWidget *page)
{
	push_page_profile(page);
	page_configure_terminal(page);
	page_configure_scrolling(page);
	page_configure_colors(page);
	pop_profile();

	page_configure_notebook(page);
	page_rename(page, page_title(page));
}

void
page_moveto(GtkWidget *op, GtkWidget *notebook)
{
	GtkWidget *label_cont;
	GtkWidget *np;
	gint nr;

	page_grab_focus(op);

	np = gtk_hbox_new(FALSE, 0);
	gtk_widget_show(np);

	label_cont = gtk_hbox_new(FALSE, 0);
	gtk_widget_show(label_cont);

	MOVE_OBJECT_INFO(op, np, PageInfo);

	gtk_widget_reparent(page_label(np),        label_cont);
	gtk_widget_reparent(page_entry(np),        label_cont);
	gtk_widget_reparent(page_close_button(np), label_cont);

	gtk_widget_reparent(page_lf_sbar(np),      np);
	gtk_widget_reparent(page_zvt(np),          np);
	gtk_widget_reparent(page_rt_sbar(np),      np);

	gtk_box_set_child_packing(GTK_BOX(label_cont),
				  page_close_button(np),
				  FALSE,
				  FALSE,
				  0,
				  GTK_PACK_START);

	gtk_box_set_child_packing(GTK_BOX(np),
			 	  page_lf_sbar(np),
				  FALSE,
				  FALSE,
				  0,
				  GTK_PACK_START);

	gtk_box_set_child_packing(GTK_BOX(np),
				  page_rt_sbar(np),
			 	  FALSE,
				  FALSE,
				  0,
				  GTK_PACK_START);

	notebook_remove_page(page_notebook(op), op);

	gtk_notebook_append_page(GTK_NOTEBOOK(notebook), np, label_cont);

	nr = gtk_notebook_page_num(GTK_NOTEBOOK(notebook), np);
	gtk_notebook_set_page(GTK_NOTEBOOK(notebook), nr);

	notebook_configure(notebook);
	page_grab_focus(np);
}

gint
page_is_lonely(GtkWidget *page)
{
	return	!page ||
		!gtk_notebook_get_nth_page(GTK_NOTEBOOK(page_notebook(page)),1);
}

void
page_remove(GtkWidget *page)
{
	page_grab_focus(page);
	notebook_remove_page(page_notebook(page), page);
}

void
page_move_left(GtkWidget *page)
{
	GtkNotebook *nb = GTK_NOTEBOOK(page_notebook(page));
	gint pgnr = gtk_notebook_page_num(nb, page);
	
	gtk_notebook_reorder_child(nb, page, pgnr-1);
}

void
page_move_right(GtkWidget *page)
{
	GtkNotebook *nb = GTK_NOTEBOOK(page_notebook(page));
	gint pgnr = gtk_notebook_page_num(nb, page);
	
	gtk_notebook_reorder_child(nb, page, pgnr+1);
}


static GtkWidget *
sbar_cont_new(void)
{
	GtkWidget *cont = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(cont), GTK_SHADOW_NONE);
	gtk_container_set_border_width(GTK_CONTAINER(cont), 0);
	return cont;
}

static GtkWidget *
page_create_helper(GtkWidget *notebook, gint profileid)
{
	GtkWidget *label_cont;
	GtkWidget *sbar;
	GtkWidget *page;
	gchar *name;
	gint nr;

	page = gtk_hbox_new(FALSE, 0);

	page_profileid(page) = profileid;

	push_profile_id(profileid);
	name = GET_STR(program.name);
	pop_profile();

	gtk_signal_connect(GTK_OBJECT(page), "destroy",
			GTK_SIGNAL_FUNC(on_page_destroy), NULL);

	label_cont = gtk_hbox_new(FALSE, 0);
	
	page_close_button(page) = gtk_button_new_with_label("X");
	gtk_box_pack_end(GTK_BOX(label_cont), page_close_button(page),
								FALSE, FALSE, 2);
	gtk_signal_connect(GTK_OBJECT(page_close_button(page)),
				"clicked",
				GTK_SIGNAL_FUNC(on_page_close_button_click),
				NULL);

	page_lf_sbar(page)  = sbar_cont_new ();
	gtk_box_pack_start(GTK_BOX(page), page_lf_sbar(page), FALSE, FALSE, 0);

	page_zvt(page) = zvt_term_new();
	gtk_box_pack_start(GTK_BOX(page), page_zvt(page), TRUE, TRUE, 0);
	gtk_signal_connect(GTK_OBJECT(page_zvt(page)), "button-press-event",
			GTK_SIGNAL_FUNC(on_page_zvt_button_press), NULL);

	gtk_drag_dest_set (page_zvt(page), 0, dnd_get_targets(&nr), nr, 0);

	gtk_signal_connect(GTK_OBJECT(page_zvt(page)), "drag_motion",
			GTK_SIGNAL_FUNC(on_page_zvt_drag_motion), NULL);
	gtk_signal_connect(GTK_OBJECT(page_zvt(page)), "drag_drop",
			GTK_SIGNAL_FUNC(on_page_zvt_drag_drop), NULL);
	gtk_signal_connect(GTK_OBJECT(page_zvt(page)), "drag_data_received",
			GTK_SIGNAL_FUNC(on_page_zvt_drag_data_received), NULL);

	sbar = gtk_vscrollbar_new(ZVT_TERM(page_zvt(page))->adjustment);
	gtk_container_add(GTK_CONTAINER(page_lf_sbar(page)), sbar);
	gtk_signal_connect(GTK_OBJECT(sbar), "button-press-event",
			GTK_SIGNAL_FUNC(on_page_sbar_button_press), NULL);

	page_rt_sbar(page) = sbar_cont_new();
	gtk_box_pack_start(GTK_BOX(page), page_rt_sbar(page), FALSE, FALSE, 0);
	gtk_widget_hide(page_rt_sbar(page));

	sbar = gtk_vscrollbar_new(ZVT_TERM(page_zvt(page))->adjustment);
	gtk_container_add(GTK_CONTAINER(page_rt_sbar(page)), sbar);
	gtk_signal_connect(GTK_OBJECT(sbar), "button-press-event",
			GTK_SIGNAL_FUNC(on_page_sbar_button_press), NULL);

	gtk_signal_connect(GTK_OBJECT(page_zvt(page)), "child_died",
			GTK_SIGNAL_FUNC(on_page_zvt_child_died), NULL);
	gtk_signal_connect(GTK_OBJECT(page_zvt(page)), "title_changed",
			GTK_SIGNAL_FUNC(on_page_zvt_title_changed), NULL);
	
	/*
	gtk_signal_connect_after(GTK_OBJECT(page_zvt(page)), "realize",
			GTK_SIGNAL_FUNC(on_page_zvt_realize),
			(gpointer)page_zvt(page));
	*/

	page_label(page) = gtk_label_new("");
	gtk_label_set_line_wrap(GTK_LABEL(page_label(page)), FALSE);
	gtk_box_pack_start(GTK_BOX(label_cont),page_label(page), TRUE, TRUE, 0);

	gtk_misc_set_alignment(GTK_MISC(page_label(page)), 0.0, 0.5);

	page_entry(page) = gtk_entry_new();
	gtk_signal_connect(GTK_OBJECT(page_entry(page)), "key_press_event",
			GTK_SIGNAL_FUNC(on_page_entry_keypress), NULL);
	gtk_signal_connect(GTK_OBJECT(page_entry(page)), "focus_out_event",
			GTK_SIGNAL_FUNC(on_page_entry_focusout), NULL);

	gtk_box_pack_start(GTK_BOX(label_cont),page_entry(page), TRUE, TRUE, 0);

	gtk_widget_show_all(page);
	gtk_widget_show_all(label_cont);

	gtk_widget_hide(page_entry(page));

	gtk_notebook_append_page(GTK_NOTEBOOK(notebook), page, label_cont);

	notebook_configure(notebook);

	nr = gtk_notebook_page_num(GTK_NOTEBOOK(notebook), page);
	gtk_notebook_set_page(GTK_NOTEBOOK(notebook), nr);

	page_rename(page, name);
	g_free(name);

	return page;
}

static gboolean
page_start_command(GtkWidget *page)
{
	gint  forked;
	gchar **argv;
	gchar *command;
	gchar *type;
	gchar *term;

	push_page_profile(page);
	command = GET_STR(program.command);
	type = GET_STR(terminal.type);
	term = g_strdup_printf("TERM=%s", type);
	g_free(type);
	pop_profile();

	while ((forked = zvt_term_forkpty(ZVT_TERM(page_zvt(page)), TRUE)) == EINTR)
		;

	if (forked < 0) {
		GtkWidget *box;
		gchar message[512];

		g_snprintf(message,
			   sizeof(message) - 1,
			   _("Fork failed when trying to start '%s': %s"),
			   command,
			   g_strerror(errno));

		box = gnome_message_box_new(message,
				GNOME_MESSAGE_BOX_ERROR,
				GNOME_STOCK_BUTTON_OK,
				NULL);

		gnome_dialog_run(GNOME_DIALOG(box));
	}

	if (forked) {
		if (forked > 0) {
			page_input_tag(page) = gdk_input_add(ZVT_TERM(page_zvt(page))->vx->vt.keyfd,
					GDK_INPUT_READ,
					(GdkInputFunction)on_page_zvt_output,
					(gpointer)page_zvt(page));
		}

		g_free(command);
		g_free(term);
		return (forked > 0);
	}

	argv = program_parse_command(command);

	putenv(term);
	execvp(argv[0], argv);

	for (;;) {
		gchar error[512];
		gchar *argv[] = { "dzt", "--error", error, NULL };

		g_snprintf(error,
			   sizeof(error) - 1,
			   _("Couldn't start '%s':\n%s"),
			   command,
			   g_strerror(errno));

		execvp(argv[0], argv);
	}
}

static void
page_configure_terminal(GtkWidget *page)
{
	GdkFont *gdk_font;
	ZvtTerm *zvt = ZVT_TERM(page_zvt(page));
	gchar *word_chars = GET_STR(terminal.word_chars);
	gchar *font = GET_STR(terminal.font);
	gchar *pixmap = GET_STR(terminal.background_image);

	zvt_term_set_bell(zvt, GET_BOOL(terminal.enable_bell));
	zvt_term_set_blink(zvt, GET_BOOL(terminal.blinking_cursor));
	zvt_term_set_del_key_swap(zvt, GET_BOOL(terminal.swap_del));
	zvt_term_set_del_is_del(zvt, GET_BOOL(terminal.del_is_del));

	if (word_chars) {
		zvt_term_set_wordclass(zvt, word_chars);
		g_free(word_chars);
	}

	if (font && strlen(font)) {
		gdk_font = gdk_font_load(font);
		if (gdk_font) {
			zvt_term_set_fonts(zvt, gdk_font, gdk_font);
			gdk_font_unref(gdk_font);
		}

		g_free(font);
	}

	switch(GET_INT(terminal.background_type)) {
		default:
			g_warning("Unknown background type.\n");

		case COLOR:
			zvt_term_set_background(zvt, NULL, FALSE, 0);
			break;

		case IMAGE:
			zvt_term_set_background(zvt, pixmap, FALSE, 0);
			break;

		case SCROLLING_IMAGE:
			zvt_term_set_background(zvt, pixmap,
					FALSE, ZVT_BACKGROUND_SCROLL);
			break;

		case TRANSPARENT:
			zvt_term_set_background(zvt, NULL, TRUE, 0);
			break;

		case TRANSPARENT_SHADED:
			zvt_term_set_background(zvt, NULL,
					TRUE, ZVT_BACKGROUND_SHADED);
			break;
	}

	g_free(pixmap);
}

static void
page_configure_scrolling(GtkWidget *page)
{
	ZvtTerm   *zvt   = ZVT_TERM(page_zvt(page));
	GtkWidget *left  = page_lf_sbar(page);
	GtkWidget *right = page_rt_sbar(page);

	zvt_term_set_scrollback(zvt, GET_INT(scrolling.scroll_buffer));
	zvt_term_set_scroll_on_keystroke(zvt, GET_BOOL(scrolling.scroll_on_keystroke));
	zvt_term_set_scroll_on_output(zvt, GET_BOOL(scrolling.scroll_on_output));

	switch(GET_INT(scrolling.scrollbar_location)) {
		case LEFT:
			gtk_widget_show(left);
			gtk_widget_hide(right);
			break;

		case RIGHT:
			gtk_widget_hide(left);
			gtk_widget_show(right);
			break;

		default:
			g_warning("Unknown scrollbar placement.\n");

		case BOTH:
			gtk_widget_show(left);
			gtk_widget_show(right);
			break;

		case NONE:
			gtk_widget_hide(left);
			gtk_widget_hide(right);
			break;
	}
}

static void
page_configure_colors(GtkWidget *page)
{
	gushort cs_red[18];
	gushort cs_grn[18];
	gushort cs_blu[18];
	gushort *color;

	gboolean invert;
	gint i;

	invert = GET_BOOL(colors.invert);

	color = GET_COLOR(colors.foreground);
	if (!color)
		return;

	cs_red[16 + invert] = color[0];
	cs_grn[16 + invert] = color[1];
	cs_blu[16 + invert] = color[2];
	g_free(color);

	color = GET_COLOR(colors.background);
	cs_red[17 - invert] = color[0];
	cs_grn[17 - invert] = color[1];
	cs_blu[17 - invert] = color[2];
	g_free(color);

	for (i = 0; i < 16; i++) {
		gint indx = i;
		if (invert)
			indx += (i > 7) ? -8 : 8;

		color = GET_NTH_COLOR(colors.pal, i);
		cs_red[indx] = color[0];
		cs_grn[indx] = color[1];
		cs_blu[indx] = color[2];
		g_free(color);
	}

	zvt_term_set_color_scheme(
			ZVT_TERM(page_zvt(page)),
			cs_red,
			cs_grn,
			cs_blu);

	page_set_changed(page, page_changed(page));
}

static void
page_configure_notebook(GtkWidget *page)
{
	if (GET_BOOL(notebook.tabs_have_close_button))
		gtk_widget_show(page_close_button(page));
	else
		gtk_widget_hide(page_close_button(page));
}

static void
page_set_changed(GtkWidget *page, gboolean value)
{
	page_changed(page) = value;
	push_page_profile(page);
	if (value) {
		gchar *color_string = GET_STR(colors.change_notify);
		gchar *ptr = color_string;
		GdkColor color = {
			0,
			(gushort)strtol(ptr, &ptr, 0),
			(gushort)strtol(ptr, &ptr, 0),
			(gushort)strtol(ptr, &ptr, 0)
		};
		GtkStyle *style;

		g_free(color_string);
		if (gdk_colormap_alloc_color(gdk_colormap_get_system(), &color, FALSE, TRUE)) {
			style = gtk_style_copy(gtk_widget_get_style(page_label(page)));
			style->fg[GTK_STATE_NORMAL] = color;
			gtk_widget_set_style(page_label(page), style);
		}
	} else
		gtk_widget_set_rc_style(page_label(page));
	pop_profile();
}

static gint 
on_page_destroy(GtkWidget *page)
{
	static gint signals[] = { SIGHUP, SIGINT, SIGQUIT, SIGTERM, SIGKILL };

	if (page_restarting(page))
		gtk_timeout_remove(page_restarting(page));

	if (page_input_tag(page))
		gdk_input_remove(page_input_tag(page));

	if (page_profileid(page)) {
		gint signal;
		push_page_profile(page);
		signal = signals[GET_INT(program.kill_signal)];
		pop_profile();

		zvt_term_killchild(ZVT_TERM(page_zvt(page)), signal);
	}
	return TRUE;
}

static void
on_page_close_button_click(GtkWidget *button)
{
	page_remove(page_from_label_component(button));
}

static gint
on_page_zvt_button_press(GtkWidget *w, GdkEventButton *ev)
{
	if (ev->button != 3)
		return FALSE;

	if (!GET_BOOL(main.no_right_click_on_terminal))
		return FALSE;

	notebook_set_ignore_click(page_notebook(w));
	return TRUE;
}

static gint
on_page_sbar_button_press(GtkWidget *w, GdkEventButton *ev)
{
	if (ev->button != 3)
		return FALSE;

	menu_popup(page_notebook(w), ev->time);
	return TRUE;
}

static void
on_page_zvt_child_died(GtkWidget *w)
{
	GtkWidget *page = page_from_widget(w);
	gboolean restart;

	push_page_profile(page);
	restart = GET_BOOL(program.restart);
	pop_profile(page);

	if (restart)
		page_restarting(page) = gtk_timeout_add(100, on_page_restart_timeout, (gpointer)w);
	else
		page_remove(page);
}

static void
on_page_zvt_title_changed(GtkWidget *w, VTTITLE_TYPE type, gchar *newtitle)
{
	page_rename(page_from_widget(w), newtitle);
}

/*
static void
on_page_zvt_realize(GtkWidget *zvt)
{
	GtkWidget	*app       = gtk_widget_get_toplevel(zvt);
	GtkWidget	*page      = page_from_widget(zvt);
	GtkNotebook	*notebook  = GTK_NOTEBOOK(page_notebook(page));
	GtkWidget	*left      = page_lf_sbar(page);
	GtkWidget	*right     = page_rt_sbar(page);
	GdkGeometry	hints;

	hints.base_width = (zvt->style->klass->xthickness * 2) + 2;
	hints.base_height = (zvt->style->klass->ythickness * 2);

	if (notebook->show_tabs) {
		hints.base_height += notebook->tab_vborder;
		hints.base_height += notebook->cur_page->tab_label->allocation.height;
	}

	if (GTK_WIDGET_VISIBLE(left))
		hints.base_width += left->allocation.width;

	if (GTK_WIDGET_VISIBLE(right))
		hints.base_width += right->allocation.width;

	hints.width_inc = ZVT_TERM(zvt)->charwidth;
	hints.height_inc = ZVT_TERM(zvt)->charheight;

	hints.min_width = hints.base_width + hints.width_inc;
	hints.min_height = hints.base_height + hints.height_inc;
		
	gtk_window_set_geometry_hints(GTK_WINDOW(app), zvt, &hints,
		GDK_HINT_RESIZE_INC|GDK_HINT_MIN_SIZE|GDK_HINT_BASE_SIZE);
}
*/

static gboolean
on_page_zvt_drag_motion (GtkWidget *zvt,
     			 GdkDragContext *ctx,
			 gint x,
			 gint y,
			 guint time)
{
	gdk_drag_status(ctx, ctx->suggested_action, time);
	return TRUE;
}

static gboolean
on_page_zvt_drag_drop (GtkWidget *zvt,
		       GdkDragContext *ctx,
		       gint x,
		       gint y,
		       guint time)
{
	if (!ctx->targets) {
		gtk_drag_finish(ctx, FALSE, FALSE, time);
		return FALSE;
	}

	gtk_drag_get_data(zvt, ctx, GPOINTER_TO_INT(ctx->targets->data), time);
	return TRUE;
}

static gboolean
on_page_zvt_drag_data_received (GtkWidget *zvt,
		       		GdkDragContext *ctx,
		       		gint x,
		       		gint y,
				GtkSelectionData *data,
				guint info,
				guint time)
{
	gchar *str = dnd_decode(data, info);
	if (str) {
		gint fd = ZVT_TERM(zvt)->vx->vt.keyfd;
		write(fd, str, strlen(str));
		g_free(str);
		gtk_drag_finish(ctx, TRUE, FALSE, time);

		return TRUE;
	}

	gtk_drag_finish(ctx, FALSE, FALSE, time);
	gdk_beep();

	return TRUE;
}

static void
on_page_entry_keypress(GtkWidget *entry, GdkEventKey *event)
{
	GtkWidget *page = page_from_label_component(entry);

	switch(event->keyval) {
		case GDK_KP_Enter:
		case GDK_ISO_Enter:
		case GDK_3270_Enter:
		case GDK_Return:
			page_end_rename(page, TRUE);

		case GDK_Escape:
			notebook_grab_focus(page_notebook(page));

		default:
			break;
	}
}

static void
on_page_entry_focusout(GtkWidget *entry, GdkEventFocus *event)
{
	page_end_rename(page_from_label_component(entry), FALSE);
}

static gboolean
on_page_restart_timeout(gpointer data)
{
	GtkWidget *page = page_from_widget(GTK_WIDGET(data));

	if (!page_start_command(page)) {
		page_restarting(page) = 0;
		page_remove(page);
		return FALSE;
	}

	page_restarting(page) = 0;
	return FALSE;
}

static void
on_page_zvt_output(gpointer data,
		   gint source,
		   GdkInputFunction condition)
{
	GtkWidget *page = page_from_widget(GTK_WIDGET(data));

	if (page == notebook_current_page(page_notebook(page)))
		return;
	if (page_changed(page))
		return;

	page_set_changed(page, TRUE);
}
