/*
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
/*
 * Copyright (c)2001,2002 Adrian Drzewiecki
 */
#include "dzt.h"

static gchar *program_command_next_arg (gchar **cmd);
static gchar *program_command_get_env  (gchar **cmd);

gchar **
program_parse_command(gchar *command)
{
	gchar **argv = NULL;
	gint    argc = 0;

	while(*command) {
		gchar *arg = program_command_next_arg(&command);
		if (!arg)
			break;
		argv = g_realloc(argv, sizeof(gchar*) * ++argc);
		argv[argc - 1] = arg;
	}

	argv = g_realloc(argv, sizeof(gchar*) * ++argc);
	argv[argc - 1] = NULL;

	return argv;
}

static gchar *
program_command_next_arg(gchar **cmd)
{
	GString *arg = g_string_sized_new(32);

	gint squote = FALSE;
	gint dquote = FALSE;
	gint escape = FALSE;

	while (**cmd && isspace(**cmd))
		++*cmd;

	if (**cmd == '~') {
		g_string_append (arg, g_getenv("HOME"));
		++*cmd;
	}

	while (**cmd) {
		if (!escape && **cmd == '\\')
			escape = TRUE;

		else if (!escape && **cmd == '\'')
			squote = !squote;

		else if (!escape && !squote && **cmd == '\"')
			dquote = !dquote;

		else if (!escape && !squote && **cmd == '$') {
			gchar *var;
			++*cmd;
			var = program_command_get_env(cmd);
			if (var) {
				g_string_append(arg, var);
				g_free(var);
			}
			continue;

		} else if (!escape && !squote && !dquote && isspace(**cmd)) {
			++*cmd;
			break;
			
		} else {
			escape = FALSE;
			g_string_append_c (arg, **cmd);
		}
		++*cmd;
	}

	if (arg->len) {
		gchar *argstr = arg->str;
		g_string_free(arg, FALSE);
		return argstr;
	}

	g_string_free(arg, TRUE);
	return NULL;
}

static gchar *
program_command_get_env(gchar **cmd)
{
	GString *var = g_string_sized_new(32);
	gchar *val = NULL;

	while(**cmd && (isalpha(**cmd) || **cmd=='_')) {
		g_string_append_c(var, **cmd);
		++*cmd;
	}

	if (var->len) {
		val = g_strdup(g_getenv(var->str));
		if (!val && strcmp("SHELL", var->str)) {
			gchar *user = g_get_user_name();
			struct passwd *pent = getpwnam(user);
			val = g_strdup(pent->pw_shell);
		}
	}
	g_string_free(var, TRUE);
	return val;
}
